package ru.dbotthepony.kstarbound

import org.apache.logging.log4j.LogManager
import org.lwjgl.Version
import org.lwjgl.glfw.Callbacks.glfwFreeCallbacks
import org.lwjgl.glfw.GLFW.*
import org.lwjgl.glfw.GLFWErrorCallback
import org.lwjgl.opengl.GL46.*
import org.lwjgl.system.MemoryStack.stackPush
import org.lwjgl.system.MemoryUtil.NULL
import ru.dbotthepony.kstarbound.gl.*
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.render.Camera
import ru.dbotthepony.kstarbound.render.ChunkRenderer
import ru.dbotthepony.kstarbound.render.TileRenderer
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE_FF
import ru.dbotthepony.kstarbound.world.ChunkTile
import java.io.File
import kotlin.math.PI

private val LOGGER = LogManager.getLogger()
private const val TEST = false

var viewportWidth = 800
	private set
var viewportHeight = 600
	private set

var viewportMatrixGUI = updateViewportMatrixA()
	private set

var viewportMatrixGame = updateViewportMatrixB()
	private set

private fun updateViewportMatrixA(): Matrix4f {
	return Matrix4f.ortho(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 0.1f, 100f)
}

private fun updateViewportMatrixB(): Matrix4f {
	return Matrix4f.orthoDirect(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 1f, 100f)
}

var window = 0L
	private set

fun main() {
	LOGGER.info("Running LWJGL ${Version.getVersion()}")

	if (!TEST) {
		try {
			init()
			loop()
		} finally {
			if (window != NULL) {
				glfwFreeCallbacks(window)
				glfwDestroyWindow(window)
			}

			glfwTerminate()
			glfwSetErrorCallback(null)?.free()
		}
	} else {
		Starbound.addFilePath(File("./unpacked_assets/"))
		Starbound.loadTileDefinition("alienrock")
	}
}

private fun init() {
	GLFWErrorCallback.create { error, description ->
		LOGGER.error("LWJGL error {}: {}", error, description)
	}.set()

	check(glfwInit()) { "Unable to initialize GLFW" }

	glfwDefaultWindowHints()

	glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE)
	glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 6)
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE)

	window = glfwCreateWindow(viewportWidth, viewportHeight, "LWJGL Window!", NULL, NULL)
	require(window != NULL) { "Unable to create GLFW window" }

	glfwSetKeyCallback(window) { window, key, scancode, action, mods ->
		if (key == GLFW_KEY_ESCAPE || key == GLFW_RELEASE) {
			glfwSetWindowShouldClose(window, true)
		}
	}

	glfwSetFramebufferSizeCallback(window) { _, w, h ->
		viewportWidth = w
		viewportHeight = h
		viewportMatrixGUI = updateViewportMatrixA()
		viewportMatrixGame = updateViewportMatrixB()
		glViewport(0, 0, w, h)
	}

	val stack = stackPush()

	try {
		val pWidth = stack.mallocInt(1)
		val pHeight = stack.mallocInt(1)

		glfwGetWindowSize(window, pWidth, pHeight)

		val vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor())!!

		glfwSetWindowPos(
			window,
			(vidmode.width() - pWidth[0]) / 2,
			(vidmode.height() - pHeight[0]) / 2
		)
	} finally {
		stack.close()
	}

	glfwMakeContextCurrent(window)

	// vsync
	glfwSwapInterval(1)

	glfwShowWindow(window)

	Starbound.addFilePath(File("./unpacked_assets/"))
	Starbound.loadTileMaterials()
}

private fun loop() {
	val state = GLStateTracker()
	val camera = Camera()

	// Set the clear color
	glClearColor(0.75f, 0.75f, 0.75f, 0.75f)

	state.blend = true
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

	val chunk = Starbound.world.getOrMakeChunk(Vector2i(2, 2))

	var x = 0
	var y = 0

	for (tile in Starbound.tilesAccess.values) {
		chunk[x, y + 1] = (tile)
		chunk[x++, y] = (tile)
		chunk[x, y + 1] = (tile)
		chunk[x++, y] = (tile)
		chunk[x, y + 1] = (tile)
		chunk[x++, y] = (tile)
		chunk[x, y + 1] = (tile)
		chunk[x++, y] = (tile)
		chunk[x, y + 1] = (tile)
		chunk[x++, y] = (tile)
		chunk[x, y + 1] = (tile)

		if (x >= 32) {
			x = 0
			y += 2
		}
	}

	val chunkRenderer = ChunkRenderer(state, chunk)
	chunkRenderer.tesselateStatic()
	chunkRenderer.uploadStatic()

	// Run the rendering loop until the user has attempted to close
	// the window or has pressed the ESCAPE key.
	while (!glfwWindowShouldClose(window)) {
		glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT) // clear the framebuffer

		state.matrixStack.clear(viewportMatrixGame.toMutableMatrix())

		state.matrixStack.push().scale(x = 20f, y = 20f).translateWithScale(0f, 0f)
		chunkRenderer.render()

		glfwSwapBuffers(window) // swap the color buffers

		// Poll for window events. The key callback above will only be
		// invoked during this call.
		glfwPollEvents()
	}
}
