package ru.dbotthepony.kstarbound.render

import ru.dbotthepony.kstarbound.gl.*
import ru.dbotthepony.kstarbound.math.FloatMatrix
import ru.dbotthepony.kstarbound.world.Chunk
import kotlin.collections.ArrayList
import kotlin.collections.HashMap

class ChunkRenderer(val state: GLStateTracker, val chunk: Chunk) : AutoCloseable {
	private val layers = TileLayerList()
	private val bakedMeshes = ArrayList<BakedStaticMesh>()
	private val unloadableBakedMeshes = ArrayList<BakedStaticMesh>()

	/**
	 * Тесселирует "статичную" геометрию в builders (к примеру тайлы).
	 *
	 * Может быть вызван вне рендер потока (ибо в любом случае он требует некой "стаитичности" данных в чанке)
	 * но только если до этого был вызыван loadRenderers() и геометрия чанка не поменялась
	 */
	fun tesselateStatic() {
		if (state.isSameThread()) {
			for (mesh in bakedMeshes) {
				mesh.close()
			}

			bakedMeshes.clear()
		} else {
			unloadableBakedMeshes.addAll(bakedMeshes)
			bakedMeshes.clear()
		}

		layers.clear()

		// TODO: Синхронизация (ибо обновления игровой логики будут в потоке вне рендер потока)
		for ((pos, tile) in chunk.posToTile) {
			if (tile != null) {
				val renderer = state.tileRenderers.get(tile.def.materialName)
				renderer.tesselate(chunk, layers, pos)
			}
		}
	}

	/**
	 * Принудительно подгружает в GLStateTracker все необходимые рендереры (ибо им нужны текстуры и прочее)
	 *
	 * Вызывается перед tesselateStatic()
	 */
	fun loadRenderers() {
		unloadUnused()

		// TODO: Синхронизация (ибо обновления игровой логики будут в потоке вне рендер потока)
		for ((pos, tile) in chunk.posToTile) {
			if (tile != null) {
				state.tileRenderers.get(tile.def.materialName)
			}
		}
	}

	private fun unloadUnused() {
		if (unloadableBakedMeshes.size != 0) {
			for (baked in unloadableBakedMeshes) {
				baked.close()
			}

			unloadableBakedMeshes.clear()
		}
	}

	fun uploadStatic(clear: Boolean = true) {
		unloadUnused()

		for ((baked, builder) in layers.buildList()) {
			bakedMeshes.add(BakedStaticMesh(baked, builder))
		}

		if (clear) {
			layers.clear()
		}
	}

	fun render(transform: FloatMatrix<*> = state.matrixStack.last) {
		unloadUnused()

		for (mesh in bakedMeshes) {
			mesh.render(transform)
		}
	}

	override fun close() {
		for (mesh in bakedMeshes) {
			mesh.close()
		}

		for (mesh in unloadableBakedMeshes) {
			mesh.close()
		}
	}
}
